@@ -1,5 +1,6 @@
 /* Calendar.java --
-   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2004, 2005  Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2004, 2005, 2006, 2007
+   Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -15,8 +16,8 @@ General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU Classpath; see the file COPYING.  If not, write to the
-Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-02111-1307 USA.
+Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+02110-1301 USA.
 
 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
@@ -42,9 +43,12 @@ import java.io.IOException;
 import java.io.ObjectInputStream;
 import java.io.ObjectOutputStream;
 import java.io.Serializable;
+
 import java.lang.reflect.Constructor;
 import java.lang.reflect.InvocationTargetException;
 
+import java.text.DateFormatSymbols;
+
 /**
  * This class is an abstract base class for Calendars, which can be
  * used to convert between <code>Date</code> objects and a set of
@@ -98,12 +102,27 @@ day_of_week + week_of_year</pre>
  * specific field by one, propagating overflows), or
  * <code>add</code>ing/substracting a fixed amount to a field.
  *
+ * @author Aaron M. Renn (arenn@urbanophile.com)
+ * @author Jochen Hoenicke (Jochen.Hoenicke@Informatik.Uni-Oldenburg.de)
+ * @author Warren Levy (warrenl@cygnus.com)
+ * @author Jeff Sturm (jsturm@one-point.com)
+ * @author Tom Tromey (tromey@redhat.com)
+ * @author Bryce McKinlay (mckinlay@redhat.com)
+ * @author Ingo Proetel (proetel@aicas.com)
+ * @author Jerry Quinn (jlquinn@optonline.net)
+ * @author Jeroen Frijters (jeroen@frijters.net)
+ * @author Noa Resare (noa@resare.com)
+ * @author Sven de Marothy (sven@physto.se)
+ * @author David Gilbert (david.gilbert@object-refinery.com)
+ * @author Olivier Jolly (olivier.jolly@pcedev.com)
+ * @author Andrew John Hughes (gnu_andrew@member.fsf.org)
  * @see Date
  * @see GregorianCalendar
  * @see TimeZone
  * @see java.text.DateFormat
  */
-public abstract class Calendar implements Serializable, Cloneable
+public abstract class Calendar
+  implements Serializable, Cloneable, Comparable<Calendar>
 {
   /**
    * Constant representing the era time field.
@@ -323,6 +342,34 @@ public abstract class Calendar implements Serializable, Cloneable
    */
   public static final int PM = 1;
 
+  /**
+   * A style specifier for {@link #getDisplayNames(int,int,Locale)}
+   * stating that names should be returned in both long and short variants.
+   *
+   * @since 1.6
+   * @see #SHORT
+   * @see #LONG
+   */
+  public static final int ALL_STYLES = 0;
+
+  /**
+   * A style specifier for {@link #getDisplayName(int,int,Locale)}
+   * and {@link #getDisplayNames(int,int,Locale)} stating that names
+   * should be returned in their short variant if applicable.
+   *
+   * @since 1.6
+   */
+  public static final int SHORT = 1;
+
+  /**
+   * A style specifier for {@link #getDisplayName(int,int,Locale)}
+   * and {@link #getDisplayNames(int,int,Locale)} stating that names
+   * should be returned in their long variant if applicable.
+   *
+   * @since 1.6
+   */
+  public static final int LONG = 2;
+
   /**
    * The time fields.  The array is indexed by the constants YEAR to
    * DST_OFFSET.
@@ -460,6 +507,8 @@ public abstract class Calendar implements Serializable, Cloneable
   /**
    * Creates a calendar representing the actual time, using the default
    * time zone and locale.
+   * 
+   * @return The new calendar.
    */
   public static synchronized Calendar getInstance()
   {
@@ -469,7 +518,12 @@ public abstract class Calendar implements Serializable, Cloneable
   /**
    * Creates a calendar representing the actual time, using the given
    * time zone and the default locale.
-   * @param zone a time zone.
+   * 
+   * @param zone a time zone (<code>null</code> not permitted).
+   * 
+   * @return The new calendar.
+   * 
+   * @throws NullPointerException if <code>zone</code> is <code>null</code>.
    */
   public static synchronized Calendar getInstance(TimeZone zone)
   {
@@ -479,7 +533,12 @@ public abstract class Calendar implements Serializable, Cloneable
   /**
    * Creates a calendar representing the actual time, using the default
    * time zone and the given locale.
-   * @param locale a locale.
+   * 
+   * @param locale a locale (<code>null</code> not permitted).
+   * 
+   * @return The new calendar.
+   * 
+   * @throws NullPointerException if <code>locale</code> is <code>null</code>.
    */
   public static synchronized Calendar getInstance(Locale locale)
   {
@@ -490,7 +549,7 @@ public abstract class Calendar implements Serializable, Cloneable
    * Cache of locale->calendar-class mappings. This avoids having to do a ResourceBundle
    * lookup for every getInstance call.
    */
-  private static HashMap cache = new HashMap();
+  private static HashMap<Locale,Class> cache = new HashMap<Locale,Class>();
 
   /** Preset argument types for calendar-class constructor lookup.  */
   private static Class[] ctorArgTypes = new Class[]
@@ -501,12 +560,18 @@ public abstract class Calendar implements Serializable, Cloneable
   /**
    * Creates a calendar representing the actual time, using the given
    * time zone and locale.
-   * @param zone a time zone.
-   * @param locale a locale.
+   * 
+   * @param zone a time zone (<code>null</code> not permitted).
+   * @param locale a locale (<code>null</code> not permitted).
+   * 
+   * @return The new calendar.
+   * 
+   * @throws NullPointerException if <code>zone</code> or <code>locale</code>
+   *     is <code>null</code>.
    */
   public static synchronized Calendar getInstance(TimeZone zone, Locale locale)
   {
-    Class calendarClass = (Class) cache.get(locale);
+    Class calendarClass = cache.get(locale);
     Throwable exception = null;
 
     try
@@ -600,6 +665,10 @@ public abstract class Calendar implements Serializable, Cloneable
   /**
    * Sets this Calendar's time to the given Date.  All time fields
    * are invalidated by this method.
+   * 
+   * @param date  the date (<code>null</code> not permitted).
+   * 
+   * @throws NullPointerException if <code>date</code> is <code>null</code>.
    */
   public final void setTime(Date date)
   {
@@ -860,6 +929,7 @@ public abstract class Calendar implements Serializable, Cloneable
                          1, 1970, JANUARY, 1, 1, 1, 1, THURSDAY, 1, AM, 0, 0, 0,
                          0, 0, zone.getRawOffset(), 0
                        };
+    complete();
     isTimeSet = false;
     areFieldsSet = false;
     isSet[field] = false;
@@ -880,7 +950,6 @@ public abstract class Calendar implements Serializable, Cloneable
 
   /**
    * Fills any unset fields in the time field list
-   * @return true if the specified field has a value.
    */
   protected void complete()
   {
@@ -898,8 +967,19 @@ public abstract class Calendar implements Serializable, Cloneable
    */
   public boolean equals(Object o)
   {
-    return (o instanceof Calendar)
-           && getTimeInMillis() == ((Calendar) o).getTimeInMillis();
+    if (! (o instanceof Calendar))
+      return false;
+    Calendar cal = (Calendar) o;
+    if (getTimeInMillis() == ((Calendar) o).getTimeInMillis()
+        && cal.getFirstDayOfWeek() == getFirstDayOfWeek()
+        && cal.isLenient() == isLenient()
+        && cal.getMinimalDaysInFirstWeek() == getMinimalDaysInFirstWeek())
+      {
+        TimeZone self = getTimeZone();
+        TimeZone oth = cal.getTimeZone();
+        return self == null ? oth == null : self.equals(oth);
+      }
+    return false;
   }
 
   /**
@@ -910,7 +990,13 @@ public abstract class Calendar implements Serializable, Cloneable
   public int hashCode()
   {
     long time = getTimeInMillis();
-    return (int) ((time & 0xffffffffL) ^ (time >> 32));
+    int val = (int) ((time & 0xffffffffL) ^ (time >> 32));
+    val += (getFirstDayOfWeek() + (isLenient() ? 1230 : 1237)
+            + getMinimalDaysInFirstWeek());
+    TimeZone self = getTimeZone();
+    if (self != null)
+      val ^= self.hashCode();
+    return val;
   }
 
   /**
@@ -1004,6 +1090,8 @@ public abstract class Calendar implements Serializable, Cloneable
   public void setTimeZone(TimeZone zone)
   {
     this.zone = zone;
+    computeTime();
+    computeFields();
   }
 
   /**
@@ -1159,6 +1247,31 @@ public abstract class Calendar implements Serializable, Cloneable
     return max;
   }
 
+  /**
+   * Compares the time of two calendar instances.
+   * @param calendar the calendar to which the time should be compared.
+   * @return 0 if the two calendars are set to the same time, 
+   * less than 0 if the time of this calendar is before that of 
+   * <code>cal</code>, or more than 0 if the time of this calendar is after
+   * that of <code>cal</code>.
+   *
+   * @param cal the calendar to compare this instance with.
+   * @throws NullPointerException if <code>cal</code> is null.
+   * @throws IllegalArgumentException if either calendar has fields set to 
+   * invalid values.
+   * @since 1.5
+   */
+  public int compareTo(Calendar cal)
+  {
+    long t1 = getTimeInMillis();
+    long t2 = cal.getTimeInMillis();
+    if(t1 == t2)
+      return 0;
+    if(t1 > t2)
+      return 1;
+    return -1;
+  }
+
   /**
    * Return a clone of this object.
    */
@@ -1258,4 +1371,205 @@ public abstract class Calendar implements Serializable, Cloneable
 	areFieldsSet = false;
       }
   }
+
+  /**
+   * Returns a localised textual representation of the current value
+   * of the given field using the specified style.  If there is no
+   * applicable textual representation (e.g. the field has a numeric
+   * value), then <code>null</code> is returned.  If one does exist,
+   * then the value is obtained from {@link #get(int)} and converted
+   * appropriately.  For example, if the <code>MONTH</code> field is
+   * requested, then <code>get(MONTH)</code> is called.  This is then
+   * converted to a textual representation based on its value and
+   * the style requested; if the <code>LONG</code> style is requested
+   * and the returned value is <code>11</code> from a
+   * {@link GregorianCalendar} implementation, then <code>"December"</code>
+   * is returned.  By default, a textual representation is available
+   * for all fields which have an applicable value obtainable from
+   * {@link java.text.DateFormatSymbols}.
+   *
+   * @param field the calendar field whose textual representation should
+   *              be obtained.
+   * @param style the style to use; either {@link #LONG} or {@link #SHORT}.
+   * @param locale the locale to use for translation.
+   * @return the textual representation of the given field in the specified
+   *         style, or <code>null</code> if none is applicable.
+   * @throws IllegalArgumentException if <code>field</code> or <code>style</code>
+   *                                  or invalid, or the calendar is non-lenient
+   *                                  and has invalid values.
+   * @throws NullPointerException if <code>locale</code> is <code>null</code>.
+   * @since 1.6
+   */
+  public String getDisplayName(int field, int style, Locale locale)
+  {
+    if (field < 0 || field >= FIELD_COUNT)
+      throw new IllegalArgumentException("The field value, " + field +
+					 ", is invalid.");
+    if (style != SHORT && style != LONG)
+      throw new IllegalArgumentException("The style must be either " +
+					 "short or long.");
+    if (field == YEAR || field == WEEK_OF_YEAR ||
+	field == WEEK_OF_MONTH || field == DAY_OF_MONTH ||
+	field == DAY_OF_YEAR || field == DAY_OF_WEEK_IN_MONTH ||
+	field == HOUR || field == HOUR_OF_DAY || field == MINUTE ||
+	field == SECOND || field == MILLISECOND)
+      return null;
+
+    int value = get(field);
+    DateFormatSymbols syms = DateFormatSymbols.getInstance(locale);
+    if (field == ERA)
+      return syms.getEras()[value];
+    if (field == MONTH)
+      if (style == LONG)
+	return syms.getMonths()[value];
+      else 
+	return syms.getShortMonths()[value];
+    if (field == DAY_OF_WEEK)
+      if (style == LONG)
+	return syms.getWeekdays()[value];
+      else
+	return syms.getShortWeekdays()[value];
+    if (field == AM_PM)
+      return syms.getAmPmStrings()[value];
+    if (field == ZONE_OFFSET)
+      if (style == LONG)
+	return syms.getZoneStrings()[value][1];
+      else
+	return syms.getZoneStrings()[value][2];
+    if (field == DST_OFFSET)
+      if (style == LONG)
+	return syms.getZoneStrings()[value][3];
+      else
+	return syms.getZoneStrings()[value][4];
+
+    throw new InternalError("Failed to resolve field " + field +
+			    " with style " + style + " for locale " +
+			    locale);
+  }
+
+  /**
+   * Returns a map linking all specified textual representations
+   * of the given field to their numerical values.  The textual
+   * representations included are determined by the specified
+   * style and locale.  For example, if the style <code>LONG</code>
+   * is specified and the German locale, then the map will
+   * contain "Montag" to {@link #MONDAY}, "Dienstag" to
+   * {@link #TUESDAY}, "Mittwoch" to {@link #WEDNESDAY} and
+   * so on.  The default implementation uses the values returned
+   * by {@link DateFormatSymbols} so, for example, the style
+   * {@link #ALL_STYLES} and the field {@link #MONTH} will return
+   * a map filled with the values returned from
+   * {@link DateFormatSymbols#getMonths()} and
+   * {@link DateFormatSymbols#getShortMonths()}.  If there are
+   * no textual representations for a given field (usually because
+   * it is purely numeric, such as the year in the
+   * {@link GregorianCalendar}), <code>null</code> is returned.
+   *
+   * @param field the calendar field whose textual representation should
+   *              be obtained.
+   * @param style the style to use; either {@link #LONG}, {@link #SHORT}
+   *              or {@link ALL_STYLES}.
+   * @param locale the locale to use for translation.
+   * @return a map of the textual representations of the given field in the
+   *         specified style to their numeric values, or <code>null</code>
+   *         if none is applicable.
+   * @throws IllegalArgumentException if <code>field</code> or <code>style</code>
+   *                                  or invalid, or the calendar is non-lenient
+   *                                  and has invalid values.
+   * @throws NullPointerException if <code>locale</code> is <code>null</code>.
+   * @since 1.6
+   */
+  public Map<String,Integer> getDisplayNames(int field, int style, Locale locale)
+  {
+    if (field < 0 || field >= FIELD_COUNT)
+      throw new IllegalArgumentException("The field value, " + field +
+					 ", is invalid.");
+    if (style != SHORT && style != LONG && style != ALL_STYLES)
+      throw new IllegalArgumentException("The style must be either " +
+					 "short, long or all styles.");
+    if (field == YEAR || field == WEEK_OF_YEAR ||
+	field == WEEK_OF_MONTH || field == DAY_OF_MONTH ||
+	field == DAY_OF_YEAR || field == DAY_OF_WEEK_IN_MONTH ||
+	field == HOUR || field == HOUR_OF_DAY || field == MINUTE ||
+	field == SECOND || field == MILLISECOND)
+      return null;
+
+    DateFormatSymbols syms = DateFormatSymbols.getInstance(locale);
+    Map<String,Integer> map = new HashMap<String,Integer>();
+    if (field == ERA)
+      {
+	String[] eras = syms.getEras();
+	for (int a = 0; a < eras.length; ++a)
+	  map.put(eras[a], a);
+	return map;
+      }
+    if (field == MONTH)
+      {
+	if (style == LONG || style == ALL_STYLES)
+	  {
+	    String[] months = syms.getMonths();
+	    for (int a = 0; a < months.length; ++a)
+	      map.put(months[a], a);
+	  }
+	if (style == SHORT || style == ALL_STYLES)
+	  {
+	    String[] months = syms.getShortMonths();
+	    for (int a = 0; a < months.length; ++a)
+	      map.put(months[a], a);
+	  }
+	return map;
+      }
+    if (field == DAY_OF_WEEK)
+      {
+	if (style == LONG || style == ALL_STYLES)
+	  {
+	    String[] weekdays = syms.getWeekdays();
+	    for (int a = SUNDAY; a < weekdays.length; ++a)
+	      map.put(weekdays[a], a);
+	  }
+	if (style == SHORT || style == ALL_STYLES)
+	  {
+	    String[] weekdays = syms.getShortWeekdays();
+	    for (int a = SUNDAY; a < weekdays.length; ++a)
+	      map.put(weekdays[a], a);
+	  }
+	return map;
+      }
+    if (field == AM_PM)
+      {
+	String[] ampms = syms.getAmPmStrings();
+	for (int a = 0; a < ampms.length; ++a)
+	  map.put(ampms[a], a);
+	return map;
+      }
+    if (field == ZONE_OFFSET)
+      {
+	String[][] zones = syms.getZoneStrings();
+	for (int a = 0; a < zones.length; ++a)
+	  {
+	    if (style == LONG || style == ALL_STYLES) 
+	      map.put(zones[a][1], a);
+	    if (style == SHORT || style == ALL_STYLES)
+	      map.put(zones[a][2], a);
+	  }
+	return map;
+      }
+    if (field == DST_OFFSET)
+      {
+	String[][] zones = syms.getZoneStrings();
+	for (int a = 0; a < zones.length; ++a)
+	  {
+	    if (style == LONG || style == ALL_STYLES) 
+	      map.put(zones[a][3], a);
+	    if (style == SHORT || style == ALL_STYLES)
+	      map.put(zones[a][4], a);
+	  }
+	return map;
+      }
+    
+    throw new InternalError("Failed to resolve field " + field +
+			    " with style " + style + " for locale " +
+			    locale);
+  }
+
 }
